@@ -21,8 +21,8 @@ import java.nio.file.Path;
 import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.Collection;
+import java.util.Collections;
 import java.util.HashSet;
-import java.util.Iterator;
 import java.util.List;
 import java.util.Set;
 import java.util.concurrent.atomic.AtomicBoolean;
@@ -30,7 +30,6 @@ import org.apache.lucene.codecs.simpletext.SimpleTextCodec;
 import org.apache.lucene.document.Document;
 import org.apache.lucene.document.Field;
 import org.apache.lucene.index.IndexWriterConfig.OpenMode;
-import org.apache.lucene.store.AlreadyClosedException;
 import org.apache.lucene.store.Directory;
 import org.apache.lucene.store.IOContext;
 import org.apache.lucene.store.IndexInput;
@@ -140,7 +139,7 @@ public class TestIndexFileDeleter extends LuceneTestCase {
             ? new String[] {"_3.scf"}
             : new String[] {"_3.cfs", "_3.cfe"};
     for (String f : cfsFiles3) {
-      assertTrue(!slowFileExists(dir, f));
+      assertFalse(slowFileExists(dir, f));
     }
 
     String[] cfsFiles1 =
@@ -185,22 +184,14 @@ public class TestIndexFileDeleter extends LuceneTestCase {
     Set<String> set2 = new HashSet<>();
     Set<String> extra = new HashSet<>();
 
-    for (int x = 0; x < files1.length; x++) {
-      set1.add(files1[x]);
-    }
-    for (int x = 0; x < files2.length; x++) {
-      set2.add(files2[x]);
-    }
-    Iterator<String> i1 = set1.iterator();
-    while (i1.hasNext()) {
-      String o = i1.next();
+    Collections.addAll(set1, files1);
+    Collections.addAll(set2, files2);
+    for (String o : set1) {
       if (!set2.contains(o)) {
         extra.add(o);
       }
     }
-    Iterator<String> i2 = set2.iterator();
-    while (i2.hasNext()) {
-      String o = i2.next();
+    for (String o : set2) {
       if (!set1.contains(o)) {
         extra.add(o);
       }
@@ -459,8 +450,7 @@ public class TestIndexFileDeleter extends LuceneTestCase {
               // suppress only FakeIOException:
               if (exc instanceof RuntimeException && exc.getMessage().equals("fake fail")) {
                 // ok to ignore
-              } else if ((exc instanceof AlreadyClosedException
-                      || exc instanceof IllegalStateException)
+              } else if (exc instanceof IllegalStateException
                   && exc.getCause() != null
                   && "fake fail".equals(exc.getCause().getMessage())) {
                 // also ok to ignore
