@@ -17,7 +17,7 @@
 package org.apache.lucene.demo.facet;
 
 import static org.apache.lucene.facet.FacetsConfig.DEFAULT_INDEX_FIELD_NAME;
-import static org.apache.lucene.sandbox.facet.ComparableUtils.byAggregatedValue;
+import static org.apache.lucene.sandbox.facet.utils.ComparableUtils.byAggregatedValue;
 
 import java.io.IOException;
 import java.util.ArrayList;
@@ -42,7 +42,6 @@ import org.apache.lucene.index.DirectoryReader;
 import org.apache.lucene.index.IndexWriter;
 import org.apache.lucene.index.IndexWriterConfig;
 import org.apache.lucene.index.IndexWriterConfig.OpenMode;
-import org.apache.lucene.sandbox.facet.ComparableUtils;
 import org.apache.lucene.sandbox.facet.FacetFieldCollectorManager;
 import org.apache.lucene.sandbox.facet.cutters.TaxonomyFacetsCutter;
 import org.apache.lucene.sandbox.facet.cutters.ranges.LongRangeFacetCutter;
@@ -56,6 +55,12 @@ import org.apache.lucene.sandbox.facet.recorders.CountFacetRecorder;
 import org.apache.lucene.sandbox.facet.recorders.LongAggregationsFacetRecorder;
 import org.apache.lucene.sandbox.facet.recorders.MultiFacetsRecorder;
 import org.apache.lucene.sandbox.facet.recorders.Reducer;
+import org.apache.lucene.sandbox.facet.utils.ComparableUtils;
+import org.apache.lucene.sandbox.facet.utils.DrillSidewaysFacetOrchestrator;
+import org.apache.lucene.sandbox.facet.utils.FacetBuilder;
+import org.apache.lucene.sandbox.facet.utils.FacetOrchestrator;
+import org.apache.lucene.sandbox.facet.utils.RangeFacetBuilderFactory;
+import org.apache.lucene.sandbox.facet.utils.TaxonomyFacetBuilder;
 import org.apache.lucene.search.DoubleValuesSource;
 import org.apache.lucene.search.IndexSearcher;
 import org.apache.lucene.search.LongValuesSource;
@@ -76,6 +81,7 @@ public class SandboxFacetsExample {
 
   private SandboxFacetsExample() {
     config.setHierarchical("Publish Date", true);
+    config.setHierarchical("Author", false);
   }
 
   /** Build the example index. */
@@ -130,6 +136,88 @@ public class SandboxFacetsExample {
     IOUtils.close(indexWriter, taxoWriter);
   }
 
+  /**
+   * Example for {@link FacetBuilder} usage - simple API that provides results in a format very
+   * similar to classic facets module. It doesn't give all flexibility available with {@link
+   * org.apache.lucene.sandbox.facet.cutters.FacetCutter} and {@link
+   * org.apache.lucene.sandbox.facet.recorders.FacetRecorder} though, see below for lower level API
+   * usage examples.
+   */
+  private List<FacetResult> simpleFacetsWithSearch() throws IOException {
+    //// init readers and searcher
+    DirectoryReader indexReader = DirectoryReader.open(indexDir);
+    IndexSearcher searcher = new IndexSearcher(indexReader);
+    TaxonomyReader taxoReader = new DirectoryTaxonomyReader(taxoDir);
+
+    //// build facets requests
+    FacetBuilder authorFacetBuilder =
+        new TaxonomyFacetBuilder(config, taxoReader, "Author").withTopN(10);
+    FacetBuilder priceFacetBuilder =
+        RangeFacetBuilderFactory.forLongRanges(
+            "Price",
+            new LongRange("0-10", 0, true, 10, true),
+            new LongRange("10-20", 10, true, 20, true));
+
+    //// Main hits collector
+    TopScoreDocCollectorManager hitsCollectorManager =
+        new TopScoreDocCollectorManager(2, Integer.MAX_VALUE);
+
+    //// Search and collect
+    TopDocs topDocs =
+        new FacetOrchestrator()
+            .addBuilder(authorFacetBuilder)
+            .addBuilder(priceFacetBuilder)
+            .collect(new MatchAllDocsQuery(), searcher, hitsCollectorManager);
+    System.out.println(
+        "Search results: totalHits: "
+            + topDocs.totalHits
+            + ", collected hits: "
+            + topDocs.scoreDocs.length);
+
+    //// Results
+    FacetResult authorResults = authorFacetBuilder.getResult();
+    FacetResult rangeResults = priceFacetBuilder.getResult();
+
+    IOUtils.close(indexReader, taxoReader);
+
+    return List.of(authorResults, rangeResults);
+  }
+
+  /** Example for {@link FacetBuilder} usage with {@link DrillSideways}. */
+  private List<FacetResult> simpleFacetsWithDrillSideways() throws IOException {
+    //// init readers and searcher
+    DirectoryReader indexReader = DirectoryReader.open(indexDir);
+    IndexSearcher searcher = new IndexSearcher(indexReader);
+    TaxonomyReader taxoReader = new DirectoryTaxonomyReader(taxoDir);
+    DrillSideways ds = new DrillSideways(searcher, config, taxoReader);
+
+    //// build facets requests
+    FacetBuilder authorFacetBuilder =
+        new TaxonomyFacetBuilder(config, taxoReader, "Author").withTopN(10);
+    FacetBuilder priceFacetBuilder =
+        RangeFacetBuilderFactory.forLongRanges(
+            "Price",
+            new LongRange("0-10", 0, true, 10, true),
+            new LongRange("10-20", 10, true, 20, true));
+
+    //// Build query and collect
+    DrillDownQuery query = new DrillDownQuery(config);
+    query.add("Author", "Lisa");
+
+    new DrillSidewaysFacetOrchestrator()
+        .addDrillDownBuilder(priceFacetBuilder)
+        .addDrillSidewaysBuilder("Author", authorFacetBuilder)
+        .collect(query, ds);
+
+    //// Results
+    FacetResult authorResults = authorFacetBuilder.getResult();
+    FacetResult rangeResults = priceFacetBuilder.getResult();
+
+    IOUtils.close(indexReader, taxoReader);
+
+    return List.of(authorResults, rangeResults);
+  }
+
   /** User runs a query and counts facets only without collecting the matching documents. */
   List<FacetResult> facetsOnly() throws IOException {
     //// (1) init readers and searcher
@@ -643,6 +731,18 @@ public class SandboxFacetsExample {
     return facetResults;
   }
 
+  /** Runs the simple search example. */
+  public List<FacetResult> runSimpleFacetsWithSearch() throws IOException {
+    index();
+    return simpleFacetsWithSearch();
+  }
+
+  /** Runs the simple drill sideways example. */
+  public List<FacetResult> runSimpleFacetsWithDrillSideways() throws IOException {
+    index();
+    return simpleFacetsWithDrillSideways();
+  }
+
   /** Runs the search example. */
   public List<FacetResult> runFacetOnly() throws IOException {
     index();
@@ -687,9 +787,22 @@ public class SandboxFacetsExample {
 
   /** Runs the search and drill-down examples and prints the results. */
   public static void main(String[] args) throws Exception {
+    SandboxFacetsExample example = new SandboxFacetsExample();
+
+    System.out.println("Simple facet counting example:");
+    System.out.println("---------------------------------------------");
+    for (FacetResult result : example.runSimpleFacetsWithSearch()) {
+      System.out.println(result);
+    }
+
+    System.out.println("Simple facet counting for drill sideways example:");
+    System.out.println("---------------------------------------------");
+    for (FacetResult result : example.runSimpleFacetsWithDrillSideways()) {
+      System.out.println(result);
+    }
+
     System.out.println("Facet counting example:");
     System.out.println("-----------------------");
-    SandboxFacetsExample example = new SandboxFacetsExample();
     List<FacetResult> results1 = example.runFacetOnly();
     System.out.println("Author: " + results1.get(0));
     System.out.println("Publish Date: " + results1.get(1));
