@@ -736,6 +736,7 @@ public class MemoryIndex {
         storePayloads,
         indexOptions,
         fieldType.docValuesType(),
+        fieldType.docValuesSkipIndexType(),
         -1,
         Collections.emptyMap(),
         fieldType.pointDimensionCount(),
@@ -759,9 +760,8 @@ public class MemoryIndex {
   private void storeVectorValues(Info info, IndexableField vectorField) {
     assert vectorField instanceof KnnFloatVectorField || vectorField instanceof KnnByteVectorField;
     switch (info.fieldInfo.getVectorEncoding()) {
-      case BYTE:
-        if (vectorField instanceof KnnByteVectorField) {
-          KnnByteVectorField byteVectorField = (KnnByteVectorField) vectorField;
+      case BYTE -> {
+        if (vectorField instanceof KnnByteVectorField byteVectorField) {
           if (info.byteVectorCount == 1) {
             throw new IllegalArgumentException(
                 "Only one value per field allowed for byte vector field ["
@@ -781,9 +781,9 @@ public class MemoryIndex {
             "Field ["
                 + vectorField.name()
                 + "] is not a byte vector field, but the field info is configured for byte vectors");
-      case FLOAT32:
-        if (vectorField instanceof KnnFloatVectorField) {
-          KnnFloatVectorField floatVectorField = (KnnFloatVectorField) vectorField;
+      }
+      case FLOAT32 -> {
+        if (vectorField instanceof KnnFloatVectorField floatVectorField) {
           if (info.floatVectorCount == 1) {
             throw new IllegalArgumentException(
                 "Only one value per field allowed for float vector field ["
@@ -803,9 +803,7 @@ public class MemoryIndex {
             "Field ["
                 + vectorField.name()
                 + "] is not a float vector field, but the field info is configured for float vectors");
-      default:
-        throw new UnsupportedOperationException(
-            "unknown vector encoding [" + info.fieldInfo.getVectorEncoding() + "]");
+      }
     }
   }
 
@@ -838,11 +836,12 @@ public class MemoryIndex {
           new FieldInfo(
               info.fieldInfo.name,
               info.fieldInfo.number,
-              info.fieldInfo.hasVectors(),
+              info.fieldInfo.hasTermVectors(),
               info.fieldInfo.hasPayloads(),
               info.fieldInfo.hasPayloads(),
               info.fieldInfo.getIndexOptions(),
               docValuesType,
+              DocValuesSkipIndexType.NONE,
               -1,
               info.fieldInfo.attributes(),
               info.fieldInfo.getPointDimensionCount(),
@@ -1494,7 +1493,6 @@ public class MemoryIndex {
 
       @Override
       public long nextOrd() throws IOException {
-        if (ord >= values.size()) return NO_MORE_ORDS;
         return ord++;
       }
 
@@ -1696,6 +1694,12 @@ public class MemoryIndex {
       }
     }
 
+    @Override
+    public DocValuesSkipper getDocValuesSkipper(String field) throws IOException {
+      // Skipping isn't needed on a 1-doc index.
+      return null;
+    }
+
     @Override
     public PointValues getPointValues(String fieldName) {
       Info info = fields.get(fieldName);
@@ -2133,15 +2137,6 @@ public class MemoryIndex {
       }
     }
 
-    @Override
-    public Fields getTermVectors(int docID) {
-      if (docID == 0) {
-        return memoryFields;
-      } else {
-        return null;
-      }
-    }
-
     @Override
     public TermVectors termVectors() {
       return new TermVectors() {
@@ -2168,12 +2163,6 @@ public class MemoryIndex {
       return 1;
     }
 
-    @Override
-    public void document(int docID, StoredFieldVisitor visitor) throws IOException {
-      if (DEBUG) System.err.println("MemoryIndexReader.document");
-      storedFields().document(docID, visitor);
-    }
-
     @Override
     public StoredFields storedFields() {
       return new StoredFields() {
@@ -2190,18 +2179,18 @@ public class MemoryIndex {
             }
             if (info.storedValues != null) {
               for (Object value : info.storedValues) {
-                if (value instanceof BytesRef) {
-                  visitor.binaryField(info.fieldInfo, BytesRef.deepCopyOf((BytesRef) value).bytes);
-                } else if (value instanceof Double) {
-                  visitor.doubleField(info.fieldInfo, (double) value);
-                } else if (value instanceof Float) {
-                  visitor.floatField(info.fieldInfo, (float) value);
-                } else if (value instanceof Long) {
-                  visitor.longField(info.fieldInfo, (long) value);
-                } else if (value instanceof Integer) {
-                  visitor.intField(info.fieldInfo, (int) value);
-                } else if (value instanceof String) {
-                  visitor.stringField(info.fieldInfo, (String) value);
+                if (value instanceof BytesRef bytes) {
+                  visitor.binaryField(info.fieldInfo, BytesRef.deepCopyOf(bytes).bytes);
+                } else if (value instanceof Double d) {
+                  visitor.doubleField(info.fieldInfo, d);
+                } else if (value instanceof Float f) {
+                  visitor.floatField(info.fieldInfo, f);
+                } else if (value instanceof Long l) {
+                  visitor.longField(info.fieldInfo, l);
+                } else if (value instanceof Integer i) {
+                  visitor.intField(info.fieldInfo, i);
+                } else if (value instanceof String s) {
+                  visitor.stringField(info.fieldInfo, s);
                 }
               }
             }
@@ -2296,7 +2285,6 @@ public class MemoryIndex {
 
   private static final class MemoryFloatVectorValues extends FloatVectorValues {
     private final Info info;
-    private int currentDoc = -1;
 
     MemoryFloatVectorValues(Info info) {
       this.info = info;
@@ -2313,14 +2301,19 @@ public class MemoryIndex {
     }
 
     @Override
-    public float[] vectorValue() {
-      if (currentDoc == 0) {
+    public float[] vectorValue(int ord) {
+      if (ord == 0) {
         return info.floatVectorValues[0];
       } else {
         return null;
       }
     }
 
+    @Override
+    public DocIndexIterator iterator() {
+      return createDenseIterator();
+    }
+
     @Override
     public VectorScorer scorer(float[] query) {
       if (query.length != info.fieldInfo.getVectorDimension()) {
@@ -2331,50 +2324,31 @@ public class MemoryIndex {
                 + info.fieldInfo.getVectorDimension());
       }
       MemoryFloatVectorValues vectorValues = new MemoryFloatVectorValues(info);
+      DocIndexIterator iterator = vectorValues.iterator();
       return new VectorScorer() {
         @Override
         public float score() throws IOException {
+          assert iterator.docID() == 0;
           return info.fieldInfo
               .getVectorSimilarityFunction()
-              .compare(vectorValues.vectorValue(), query);
+              .compare(vectorValues.vectorValue(0), query);
         }
 
         @Override
         public DocIdSetIterator iterator() {
-          return vectorValues;
+          return iterator;
         }
       };
     }
 
     @Override
-    public int docID() {
-      return currentDoc;
-    }
-
-    @Override
-    public int nextDoc() {
-      int doc = ++currentDoc;
-      if (doc == 0) {
-        return doc;
-      } else {
-        return NO_MORE_DOCS;
-      }
-    }
-
-    @Override
-    public int advance(int target) {
-      if (target == 0) {
-        currentDoc = target;
-        return target;
-      } else {
-        return NO_MORE_DOCS;
-      }
+    public MemoryFloatVectorValues copy() {
+      return this;
     }
   }
 
   private static final class MemoryByteVectorValues extends ByteVectorValues {
     private final Info info;
-    private int currentDoc = -1;
 
     MemoryByteVectorValues(Info info) {
       this.info = info;
@@ -2391,14 +2365,19 @@ public class MemoryIndex {
     }
 
     @Override
-    public byte[] vectorValue() {
-      if (currentDoc == 0) {
+    public byte[] vectorValue(int ord) {
+      if (ord == 0) {
         return info.byteVectorValues[0];
       } else {
         return null;
       }
     }
 
+    @Override
+    public DocIndexIterator iterator() {
+      return createDenseIterator();
+    }
+
     @Override
     public VectorScorer scorer(byte[] query) {
       if (query.length != info.fieldInfo.getVectorDimension()) {
@@ -2409,44 +2388,26 @@ public class MemoryIndex {
                 + info.fieldInfo.getVectorDimension());
       }
       MemoryByteVectorValues vectorValues = new MemoryByteVectorValues(info);
+      DocIndexIterator iterator = vectorValues.iterator();
       return new VectorScorer() {
         @Override
         public float score() {
+          assert iterator.docID() == 0;
           return info.fieldInfo
               .getVectorSimilarityFunction()
-              .compare(vectorValues.vectorValue(), query);
+              .compare(vectorValues.vectorValue(0), query);
         }
 
         @Override
         public DocIdSetIterator iterator() {
-          return vectorValues;
+          return iterator;
         }
       };
     }
 
     @Override
-    public int docID() {
-      return currentDoc;
-    }
-
-    @Override
-    public int nextDoc() {
-      int doc = ++currentDoc;
-      if (doc == 0) {
-        return doc;
-      } else {
-        return NO_MORE_DOCS;
-      }
-    }
-
-    @Override
-    public int advance(int target) {
-      if (target == 0) {
-        currentDoc = target;
-        return target;
-      } else {
-        return NO_MORE_DOCS;
-      }
+    public MemoryByteVectorValues copy() {
+      return this;
     }
   }
 }
