@@ -16,6 +16,7 @@
  */
 package org.apache.lucene.tests.index;
 
+import static java.nio.charset.StandardCharsets.UTF_8;
 import static org.apache.lucene.index.VectorSimilarityFunction.DOT_PRODUCT;
 import static org.apache.lucene.search.DocIdSetIterator.NO_MORE_DOCS;
 
@@ -45,6 +46,7 @@ import org.apache.lucene.index.ByteVectorValues;
 import org.apache.lucene.index.CheckIndex;
 import org.apache.lucene.index.CodecReader;
 import org.apache.lucene.index.DirectoryReader;
+import org.apache.lucene.index.DocValuesSkipIndexType;
 import org.apache.lucene.index.DocValuesType;
 import org.apache.lucene.index.FieldInfo;
 import org.apache.lucene.index.FieldInfos;
@@ -53,6 +55,7 @@ import org.apache.lucene.index.IndexOptions;
 import org.apache.lucene.index.IndexReader;
 import org.apache.lucene.index.IndexWriter;
 import org.apache.lucene.index.IndexWriterConfig;
+import org.apache.lucene.index.KnnVectorValues;
 import org.apache.lucene.index.LeafReader;
 import org.apache.lucene.index.LeafReaderContext;
 import org.apache.lucene.index.MergePolicy;
@@ -105,16 +108,20 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
   @Override
   protected void addRandomFields(Document doc) {
     switch (vectorEncoding) {
-      case BYTE:
-        doc.add(new KnnByteVectorField("v2", randomVector8(30), similarityFunction));
-        break;
-      default:
-      case FLOAT32:
-        doc.add(new KnnFloatVectorField("v2", randomNormalizedVector(30), similarityFunction));
-        break;
+      case BYTE -> doc.add(new KnnByteVectorField("v2", randomVector8(30), similarityFunction));
+      case FLOAT32 ->
+          doc.add(new KnnFloatVectorField("v2", randomNormalizedVector(30), similarityFunction));
     }
   }
 
+  @Override
+  protected boolean mergeIsStable() {
+    // suppress this test from base class: merges for knn graphs are not stable due to connected
+    // components
+    // logic
+    return false;
+  }
+
   private int getVectorsMaxDimensions(String fieldName) {
     return Codec.getDefault().knnVectorsFormat().getMaxDimensions(fieldName);
   }
@@ -245,9 +252,7 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
     try (var dir = newDirectory()) {
       IndexWriterConfig iwc = new IndexWriterConfig();
       Codec codec = getCodec();
-      if (codec.knnVectorsFormat() instanceof PerFieldKnnVectorsFormat) {
-        PerFieldKnnVectorsFormat perFieldKnnVectorsFormat =
-            (PerFieldKnnVectorsFormat) codec.knnVectorsFormat();
+      if (codec.knnVectorsFormat() instanceof PerFieldKnnVectorsFormat perFieldKnnVectorsFormat) {
         final KnnVectorsFormat format =
             perFieldKnnVectorsFormat.getKnnVectorsFormatForField("field");
         iwc.setCodec(
@@ -284,9 +289,7 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
     try (var dir = newDirectory()) {
       IndexWriterConfig iwc = new IndexWriterConfig();
       Codec codec = getCodec();
-      if (codec.knnVectorsFormat() instanceof PerFieldKnnVectorsFormat) {
-        PerFieldKnnVectorsFormat perFieldKnnVectorsFormat =
-            (PerFieldKnnVectorsFormat) codec.knnVectorsFormat();
+      if (codec.knnVectorsFormat() instanceof PerFieldKnnVectorsFormat perFieldKnnVectorsFormat) {
         final KnnVectorsFormat format =
             perFieldKnnVectorsFormat.getKnnVectorsFormatForField("field");
         iwc.setCodec(
@@ -386,6 +389,7 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
                       false,
                       IndexOptions.NONE,
                       DocValuesType.NONE,
+                      DocValuesSkipIndexType.NONE,
                       -1,
                       Map.of(),
                       0,
@@ -442,9 +446,10 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
         try (IndexReader reader = DirectoryReader.open(w2)) {
           LeafReader r = getOnlyLeafReader(reader);
           FloatVectorValues vectorValues = r.getFloatVectorValues(fieldName);
-          assertEquals(0, vectorValues.nextDoc());
-          assertEquals(0, vectorValues.vectorValue()[0], 0);
-          assertEquals(NO_MORE_DOCS, vectorValues.nextDoc());
+          KnnVectorValues.DocIndexIterator iterator = vectorValues.iterator();
+          assertEquals(0, iterator.nextDoc());
+          assertEquals(0, vectorValues.vectorValue(0)[0], 0);
+          assertEquals(NO_MORE_DOCS, iterator.nextDoc());
         }
       }
     }
@@ -467,9 +472,10 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
         try (IndexReader reader = DirectoryReader.open(w2)) {
           LeafReader r = getOnlyLeafReader(reader);
           FloatVectorValues vectorValues = r.getFloatVectorValues(fieldName);
-          assertNotEquals(NO_MORE_DOCS, vectorValues.nextDoc());
-          assertEquals(0, vectorValues.vectorValue()[0], 0);
-          assertEquals(NO_MORE_DOCS, vectorValues.nextDoc());
+          KnnVectorValues.DocIndexIterator iterator = vectorValues.iterator();
+          assertNotEquals(NO_MORE_DOCS, iterator.nextDoc());
+          assertEquals(0, vectorValues.vectorValue(iterator.index())[0], 0);
+          assertEquals(NO_MORE_DOCS, iterator.nextDoc());
         }
       }
     }
@@ -494,12 +500,13 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
         try (IndexReader reader = DirectoryReader.open(w2)) {
           LeafReader r = getOnlyLeafReader(reader);
           FloatVectorValues vectorValues = r.getFloatVectorValues(fieldName);
-          assertEquals(0, vectorValues.nextDoc());
+          KnnVectorValues.DocIndexIterator iterator = vectorValues.iterator();
+          assertEquals(0, iterator.nextDoc());
           // The merge order is randomized, we might get 0 first, or 1
-          float value = vectorValues.vectorValue()[0];
+          float value = vectorValues.vectorValue(0)[0];
           assertTrue(value == 0 || value == 1);
-          assertEquals(1, vectorValues.nextDoc());
-          value += vectorValues.vectorValue()[0];
+          assertEquals(1, iterator.nextDoc());
+          value += vectorValues.vectorValue(1)[0];
           assertEquals(1, value, 0);
         }
       }
@@ -749,7 +756,7 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
     }
   }
 
-  // Write vectors, one segment with with SimpleText, another with default codec, then forceMerge
+  // Write vectors, one segment with SimpleText, another with default codec, then forceMerge
   public void testDifferentCodecs2() throws Exception {
     IndexWriterConfig iwc = newIndexWriterConfig();
     iwc.setCodec(Codec.forName("SimpleText"));
@@ -857,21 +864,16 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
           String fieldName = "int" + field;
           if (random().nextInt(100) == 17) {
             switch (fieldVectorEncodings[field]) {
-              case BYTE:
-                {
-                  byte[] b = randomVector8(fieldDims[field]);
-                  doc.add(new KnnByteVectorField(fieldName, b, fieldSimilarityFunctions[field]));
-                  fieldTotals[field] += b[0];
-                  break;
-                }
-              default:
-              case FLOAT32:
-                {
-                  float[] v = randomNormalizedVector(fieldDims[field]);
-                  doc.add(new KnnFloatVectorField(fieldName, v, fieldSimilarityFunctions[field]));
-                  fieldTotals[field] += v[0];
-                  break;
-                }
+              case BYTE -> {
+                byte[] b = randomVector8(fieldDims[field]);
+                doc.add(new KnnByteVectorField(fieldName, b, fieldSimilarityFunctions[field]));
+                fieldTotals[field] += b[0];
+              }
+              case FLOAT32 -> {
+                float[] v = randomNormalizedVector(fieldDims[field]);
+                doc.add(new KnnFloatVectorField(fieldName, v, fieldSimilarityFunctions[field]));
+                fieldTotals[field] += v[0];
+              }
             }
             fieldDocCounts[field]++;
           }
@@ -884,28 +886,32 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
           double checksum = 0;
           String fieldName = "int" + field;
           switch (fieldVectorEncodings[field]) {
-            case BYTE:
+            case BYTE -> {
               for (LeafReaderContext ctx : r.leaves()) {
                 ByteVectorValues byteVectorValues = ctx.reader().getByteVectorValues(fieldName);
                 if (byteVectorValues != null) {
                   docCount += byteVectorValues.size();
-                  while (byteVectorValues.nextDoc() != NO_MORE_DOCS) {
-                    checksum += byteVectorValues.vectorValue()[0];
+                  KnnVectorValues.DocIndexIterator iterator = byteVectorValues.iterator();
+                  while (true) {
+                    if (!(iterator.nextDoc() != NO_MORE_DOCS)) break;
+                    checksum += byteVectorValues.vectorValue(iterator.index())[0];
                   }
                 }
               }
-              break;
-            default:
-            case FLOAT32:
+            }
+            case FLOAT32 -> {
               for (LeafReaderContext ctx : r.leaves()) {
                 FloatVectorValues vectorValues = ctx.reader().getFloatVectorValues(fieldName);
                 if (vectorValues != null) {
                   docCount += vectorValues.size();
-                  while (vectorValues.nextDoc() != NO_MORE_DOCS) {
-                    checksum += vectorValues.vectorValue()[0];
+                  KnnVectorValues.DocIndexIterator iterator = vectorValues.iterator();
+                  while (true) {
+                    if (!(iterator.nextDoc() != NO_MORE_DOCS)) break;
+                    checksum += vectorValues.vectorValue(iterator.index())[0];
                   }
                 }
               }
+            }
           }
           assertEquals(fieldDocCounts[field], docCount);
           // Account for quantization done when indexing fields w/BYTE encoding
@@ -960,10 +966,12 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
           assertSame(iterator, scorer.iterator());
           assertNotSame(iterator, scorer);
           // verify scorer iteration scores are valid & iteration with vectorValues is consistent
-          while (iterator.nextDoc() != NO_MORE_DOCS && vectorValues.nextDoc() != NO_MORE_DOCS) {
+          KnnVectorValues.DocIndexIterator valuesIterator = vectorValues.iterator();
+          while (iterator.nextDoc() != NO_MORE_DOCS) {
+            if (!(valuesIterator.nextDoc() != NO_MORE_DOCS)) break;
             float score = scorer.score();
             assertTrue(score >= 0f);
-            assertEquals(iterator.docID(), vectorValues.docID());
+            assertEquals(iterator.docID(), valuesIterator.docID());
           }
           // verify that a new scorer can be obtained after iteration
           VectorScorer newScorer = vectorValues.scorer(vectorToScore);
@@ -1019,10 +1027,12 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
           assertSame(iterator, scorer.iterator());
           assertNotSame(iterator, scorer);
           // verify scorer iteration scores are valid & iteration with vectorValues is consistent
-          while (iterator.nextDoc() != NO_MORE_DOCS && vectorValues.nextDoc() != NO_MORE_DOCS) {
+          KnnVectorValues.DocIndexIterator valuesIterator = vectorValues.iterator();
+          while (iterator.nextDoc() != NO_MORE_DOCS) {
+            if (!(valuesIterator.nextDoc() != NO_MORE_DOCS)) break;
             float score = scorer.score();
             assertTrue(score >= 0f);
-            assertEquals(iterator.docID(), vectorValues.docID());
+            assertEquals(iterator.docID(), valuesIterator.docID());
           }
           // verify that a new scorer can be obtained after iteration
           VectorScorer newScorer = vectorValues.scorer(vectorToScore);
@@ -1128,12 +1138,16 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
         LeafReader r = getOnlyLeafReader(reader);
         FloatVectorValues vectorValues = r.getFloatVectorValues(fieldName);
         assertEquals(3, vectorValues.size());
-        vectorValues.nextDoc();
-        assertEquals(1, vectorValues.vectorValue()[0], 0);
-        vectorValues.nextDoc();
-        assertEquals(1, vectorValues.vectorValue()[0], 0);
-        vectorValues.nextDoc();
-        assertEquals(2, vectorValues.vectorValue()[0], 0);
+        KnnVectorValues.DocIndexIterator iterator = vectorValues.iterator();
+        iterator.nextDoc();
+        assertEquals(0, iterator.index());
+        assertEquals(1, vectorValues.vectorValue(0)[0], 0);
+        iterator.nextDoc();
+        assertEquals(1, iterator.index());
+        assertEquals(1, vectorValues.vectorValue(1)[0], 0);
+        iterator.nextDoc();
+        assertEquals(2, iterator.index());
+        assertEquals(2, vectorValues.vectorValue(2)[0], 0);
       }
     }
   }
@@ -1156,13 +1170,14 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
         FloatVectorValues vectorValues = leaf.getFloatVectorValues(fieldName);
         assertEquals(2, vectorValues.dimension());
         assertEquals(3, vectorValues.size());
-        assertEquals("1", storedFields.document(vectorValues.nextDoc()).get("id"));
-        assertEquals(-1f, vectorValues.vectorValue()[0], 0);
-        assertEquals("2", storedFields.document(vectorValues.nextDoc()).get("id"));
-        assertEquals(1, vectorValues.vectorValue()[0], 0);
-        assertEquals("4", storedFields.document(vectorValues.nextDoc()).get("id"));
-        assertEquals(0, vectorValues.vectorValue()[0], 0);
-        assertEquals(NO_MORE_DOCS, vectorValues.nextDoc());
+        KnnVectorValues.DocIndexIterator iterator = vectorValues.iterator();
+        assertEquals("1", storedFields.document(iterator.nextDoc()).get("id"));
+        assertEquals(-1f, vectorValues.vectorValue(0)[0], 0);
+        assertEquals("2", storedFields.document(iterator.nextDoc()).get("id"));
+        assertEquals(1, vectorValues.vectorValue(1)[0], 0);
+        assertEquals("4", storedFields.document(iterator.nextDoc()).get("id"));
+        assertEquals(0, vectorValues.vectorValue(2)[0], 0);
+        assertEquals(NO_MORE_DOCS, iterator.nextDoc());
       }
     }
   }
@@ -1185,13 +1200,13 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
         ByteVectorValues vectorValues = leaf.getByteVectorValues(fieldName);
         assertEquals(2, vectorValues.dimension());
         assertEquals(3, vectorValues.size());
-        assertEquals("1", storedFields.document(vectorValues.nextDoc()).get("id"));
-        assertEquals(-1, vectorValues.vectorValue()[0], 0);
-        assertEquals("2", storedFields.document(vectorValues.nextDoc()).get("id"));
-        assertEquals(1, vectorValues.vectorValue()[0], 0);
-        assertEquals("4", storedFields.document(vectorValues.nextDoc()).get("id"));
-        assertEquals(0, vectorValues.vectorValue()[0], 0);
-        assertEquals(NO_MORE_DOCS, vectorValues.nextDoc());
+        assertEquals("1", storedFields.document(vectorValues.iterator().nextDoc()).get("id"));
+        assertEquals(-1, vectorValues.vectorValue(0)[0], 0);
+        assertEquals("2", storedFields.document(vectorValues.iterator().nextDoc()).get("id"));
+        assertEquals(1, vectorValues.vectorValue(1)[0], 0);
+        assertEquals("4", storedFields.document(vectorValues.iterator().nextDoc()).get("id"));
+        assertEquals(0, vectorValues.vectorValue(2)[0], 0);
+        assertEquals(NO_MORE_DOCS, vectorValues.iterator().nextDoc());
       }
     }
   }
@@ -1221,27 +1236,30 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
         FloatVectorValues vectorValues = leaf.getFloatVectorValues("field1");
         assertEquals(2, vectorValues.dimension());
         assertEquals(2, vectorValues.size());
-        vectorValues.nextDoc();
-        assertEquals(1f, vectorValues.vectorValue()[0], 0);
-        vectorValues.nextDoc();
-        assertEquals(2f, vectorValues.vectorValue()[0], 0);
-        assertEquals(NO_MORE_DOCS, vectorValues.nextDoc());
+        KnnVectorValues.DocIndexIterator iterator = vectorValues.iterator();
+        iterator.nextDoc();
+        assertEquals(1f, vectorValues.vectorValue(0)[0], 0);
+        iterator.nextDoc();
+        assertEquals(2f, vectorValues.vectorValue(1)[0], 0);
+        assertEquals(NO_MORE_DOCS, iterator.nextDoc());
 
         FloatVectorValues vectorValues2 = leaf.getFloatVectorValues("field2");
+        KnnVectorValues.DocIndexIterator it2 = vectorValues2.iterator();
         assertEquals(4, vectorValues2.dimension());
         assertEquals(2, vectorValues2.size());
-        vectorValues2.nextDoc();
-        assertEquals(2f, vectorValues2.vectorValue()[1], 0);
-        vectorValues2.nextDoc();
-        assertEquals(2f, vectorValues2.vectorValue()[1], 0);
-        assertEquals(NO_MORE_DOCS, vectorValues2.nextDoc());
+        it2.nextDoc();
+        assertEquals(2f, vectorValues2.vectorValue(0)[1], 0);
+        it2.nextDoc();
+        assertEquals(2f, vectorValues2.vectorValue(1)[1], 0);
+        assertEquals(NO_MORE_DOCS, it2.nextDoc());
 
         FloatVectorValues vectorValues3 = leaf.getFloatVectorValues("field3");
         assertEquals(4, vectorValues3.dimension());
         assertEquals(1, vectorValues3.size());
-        vectorValues3.nextDoc();
-        assertEquals(1f, vectorValues3.vectorValue()[0], 0.1);
-        assertEquals(NO_MORE_DOCS, vectorValues3.nextDoc());
+        KnnVectorValues.DocIndexIterator it3 = vectorValues3.iterator();
+        it3.nextDoc();
+        assertEquals(1f, vectorValues3.vectorValue(0)[0], 0.1);
+        assertEquals(NO_MORE_DOCS, it3.nextDoc());
       }
     }
   }
@@ -1305,13 +1323,15 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
           totalSize += vectorValues.size();
           StoredFields storedFields = ctx.reader().storedFields();
           int docId;
-          while ((docId = vectorValues.nextDoc()) != NO_MORE_DOCS) {
-            float[] v = vectorValues.vectorValue();
+          KnnVectorValues.DocIndexIterator iterator = vectorValues.iterator();
+          while (true) {
+            if (!((docId = iterator.nextDoc()) != NO_MORE_DOCS)) break;
+            float[] v = vectorValues.vectorValue(iterator.index());
             assertEquals(dimension, v.length);
             String idString = storedFields.document(docId).getField("id").stringValue();
             int id = Integer.parseInt(idString);
             if (ctx.reader().getLiveDocs() == null || ctx.reader().getLiveDocs().get(docId)) {
-              assertArrayEquals(idString, values[id], v, 0);
+              assertArrayEquals(idString + " " + docId, values[id], v, 0);
               ++valueCount;
             } else {
               ++numDeletes;
@@ -1385,8 +1405,10 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
           totalSize += vectorValues.size();
           StoredFields storedFields = ctx.reader().storedFields();
           int docId;
-          while ((docId = vectorValues.nextDoc()) != NO_MORE_DOCS) {
-            byte[] v = vectorValues.vectorValue();
+          KnnVectorValues.DocIndexIterator iterator = vectorValues.iterator();
+          while (true) {
+            if (!((docId = iterator.nextDoc()) != NO_MORE_DOCS)) break;
+            byte[] v = vectorValues.vectorValue(iterator.index());
             assertEquals(dimension, v.length);
             String idString = storedFields.document(docId).getField("id").stringValue();
             int id = Integer.parseInt(idString);
@@ -1450,8 +1472,8 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
               ctx.reader()
                   .searchNearestVectors(
                       fieldName, randomNormalizedVector(dimension), k, liveDocs, visitedLimit);
-          assertEquals(TotalHits.Relation.GREATER_THAN_OR_EQUAL_TO, results.totalHits.relation);
-          assertEquals(visitedLimit, results.totalHits.value);
+          assertEquals(TotalHits.Relation.GREATER_THAN_OR_EQUAL_TO, results.totalHits.relation());
+          assertEquals(visitedLimit, results.totalHits.value());
 
           // check the limit is not hit when it clearly exceeds the number of vectors
           k = vectorValues.size();
@@ -1460,8 +1482,8 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
               ctx.reader()
                   .searchNearestVectors(
                       fieldName, randomNormalizedVector(dimension), k, liveDocs, visitedLimit);
-          assertEquals(TotalHits.Relation.EQUAL_TO, results.totalHits.relation);
-          assertTrue(results.totalHits.value <= visitedLimit);
+          assertEquals(TotalHits.Relation.EQUAL_TO, results.totalHits.relation());
+          assertTrue(results.totalHits.value() <= visitedLimit);
         }
       }
     }
@@ -1469,7 +1491,7 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
 
   /**
    * Index random vectors, sometimes skipping documents, sometimes updating a document, sometimes
-   * merging, sometimes sorting the index, using an HNSW similarity function so as to also produce a
+   * merging, sometimes sorting the index, using an HNSW similarity function to also produce a
    * graph, and verify that the expected values can be read back consistently.
    */
   public void testRandomWithUpdatesAndGraph() throws Exception {
@@ -1505,8 +1527,10 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
           StoredFields storedFields = ctx.reader().storedFields();
           int docId;
           int numLiveDocsWithVectors = 0;
-          while ((docId = vectorValues.nextDoc()) != NO_MORE_DOCS) {
-            float[] v = vectorValues.vectorValue();
+          KnnVectorValues.DocIndexIterator iterator = vectorValues.iterator();
+          while (true) {
+            if (!((docId = iterator.nextDoc()) != NO_MORE_DOCS)) break;
+            float[] v = vectorValues.vectorValue(iterator.index());
             assertEquals(dimension, v.length);
             String idString = storedFields.document(docId).getField("id").stringValue();
             int id = Integer.parseInt(idString);
@@ -1531,7 +1555,7 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
           // assert that searchNearestVectors returns the expected number of documents,
           // in descending score order
           int size = ctx.reader().getFloatVectorValues(fieldName).size();
-          int k = random().nextInt(size / 2 + 1) + 1;
+          int k = random().nextInt(size / 10 + 1) + 1;
           if (k > numLiveDocsWithVectors) {
             k = numLiveDocsWithVectors;
           }
@@ -1659,7 +1683,9 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
       }
 
       ByteArrayOutputStream output = new ByteArrayOutputStream();
-      CheckIndex.Status status = TestUtil.checkIndex(dir, false, true, true, output);
+      CheckIndex.Status status =
+          TestUtil.checkIndex(
+              dir, CheckIndex.Level.MIN_LEVEL_FOR_INTEGRITY_CHECKS, true, true, output);
       assertEquals(1, status.segmentInfos.size());
       CheckIndex.Status.SegmentInfoStatus segStatus = status.segmentInfos.get(0);
       // total 3 vector values were indexed:
@@ -1668,7 +1694,7 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
       assertEquals(2, segStatus.vectorValuesStatus.totalKnnVectorFields);
 
       // Make sure CheckIndex in fact declares that it is testing vectors!
-      assertTrue(output.toString(IOUtils.UTF_8).contains("test: vectors..."));
+      assertTrue(output.toString(UTF_8).contains("test: vectors..."));
     }
   }
 
@@ -1711,25 +1737,27 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
           FloatVectorValues vectorValues = r.getFloatVectorValues(fieldName);
           int[] vectorDocs = new int[vectorValues.size() + 1];
           int cur = -1;
+          KnnVectorValues.DocIndexIterator iterator = vectorValues.iterator();
           while (++cur < vectorValues.size() + 1) {
-            vectorDocs[cur] = vectorValues.nextDoc();
+            vectorDocs[cur] = iterator.nextDoc();
             if (cur != 0) {
               assertTrue(vectorDocs[cur] > vectorDocs[cur - 1]);
             }
           }
           vectorValues = r.getFloatVectorValues(fieldName);
+          DocIdSetIterator iter = vectorValues.iterator();
           cur = -1;
           for (int i = 0; i < numdocs; i++) {
             // randomly advance to i
             if (random().nextInt(4) == 3) {
               while (vectorDocs[++cur] < i) {}
-              assertEquals(vectorDocs[cur], vectorValues.advance(i));
-              assertEquals(vectorDocs[cur], vectorValues.docID());
-              if (vectorValues.docID() == NO_MORE_DOCS) {
+              assertEquals(vectorDocs[cur], iter.advance(i));
+              assertEquals(vectorDocs[cur], iter.docID());
+              if (iter.docID() == NO_MORE_DOCS) {
                 break;
               }
               // make i equal to docid so that it is greater than docId in the next loop iteration
-              i = vectorValues.docID();
+              i = iter.docID();
             }
           }
         }
@@ -1755,20 +1783,16 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
         doc.add(new StoredField("id", docID));
         if (random().nextInt(4) == 3) {
           switch (vectorEncoding) {
-            case BYTE:
-              {
-                byte[] b = randomVector8(dim);
-                fieldValuesCheckSum += b[0];
-                doc.add(new KnnByteVectorField("knn_vector", b, similarityFunction));
-                break;
-              }
-            case FLOAT32:
-              {
-                float[] v = randomNormalizedVector(dim);
-                fieldValuesCheckSum += v[0];
-                doc.add(new KnnFloatVectorField("knn_vector", v, similarityFunction));
-                break;
-              }
+            case BYTE -> {
+              byte[] b = randomVector8(dim);
+              fieldValuesCheckSum += b[0];
+              doc.add(new KnnByteVectorField("knn_vector", b, similarityFunction));
+            }
+            case FLOAT32 -> {
+              float[] v = randomNormalizedVector(dim);
+              fieldValuesCheckSum += v[0];
+              doc.add(new KnnFloatVectorField("knn_vector", v, similarityFunction));
+            }
           }
           fieldDocCount++;
           fieldSumDocIDs += docID;
@@ -1784,36 +1808,49 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
         double checksum = 0;
         int docCount = 0;
         long sumDocIds = 0;
+        long sumOrdToDocIds = 0;
         switch (vectorEncoding) {
-          case BYTE:
+          case BYTE -> {
             for (LeafReaderContext ctx : r.leaves()) {
               ByteVectorValues byteVectorValues = ctx.reader().getByteVectorValues("knn_vector");
               if (byteVectorValues != null) {
                 docCount += byteVectorValues.size();
                 StoredFields storedFields = ctx.reader().storedFields();
-                while (byteVectorValues.nextDoc() != NO_MORE_DOCS) {
-                  checksum += byteVectorValues.vectorValue()[0];
-                  Document doc = storedFields.document(byteVectorValues.docID(), Set.of("id"));
+                KnnVectorValues.DocIndexIterator iter = byteVectorValues.iterator();
+                for (iter.nextDoc(); iter.docID() != NO_MORE_DOCS; iter.nextDoc()) {
+                  int ord = iter.index();
+                  checksum += byteVectorValues.vectorValue(ord)[0];
+                  Document doc = storedFields.document(iter.docID(), Set.of("id"));
                   sumDocIds += Integer.parseInt(doc.get("id"));
                 }
+                for (int ord = 0; ord < byteVectorValues.size(); ord++) {
+                  Document doc =
+                      storedFields.document(byteVectorValues.ordToDoc(ord), Set.of("id"));
+                  sumOrdToDocIds += Integer.parseInt(doc.get("id"));
+                }
               }
             }
-            break;
-          default:
-          case FLOAT32:
+          }
+          case FLOAT32 -> {
             for (LeafReaderContext ctx : r.leaves()) {
               FloatVectorValues vectorValues = ctx.reader().getFloatVectorValues("knn_vector");
               if (vectorValues != null) {
                 docCount += vectorValues.size();
                 StoredFields storedFields = ctx.reader().storedFields();
-                while (vectorValues.nextDoc() != NO_MORE_DOCS) {
-                  checksum += vectorValues.vectorValue()[0];
-                  Document doc = storedFields.document(vectorValues.docID(), Set.of("id"));
+                KnnVectorValues.DocIndexIterator iter = vectorValues.iterator();
+                for (iter.nextDoc(); iter.docID() != NO_MORE_DOCS; iter.nextDoc()) {
+                  int ord = iter.index();
+                  checksum += vectorValues.vectorValue(ord)[0];
+                  Document doc = storedFields.document(iter.docID(), Set.of("id"));
                   sumDocIds += Integer.parseInt(doc.get("id"));
                 }
+                for (int ord = 0; ord < vectorValues.size(); ord++) {
+                  Document doc = storedFields.document(vectorValues.ordToDoc(ord), Set.of("id"));
+                  sumOrdToDocIds += Integer.parseInt(doc.get("id"));
+                }
               }
             }
-            break;
+          }
         }
         assertEquals(
             "encoding=" + vectorEncoding,
@@ -1822,6 +1859,7 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
             vectorEncoding == VectorEncoding.BYTE ? numDocs * 0.2 : 1e-5);
         assertEquals(fieldDocCount, docCount);
         assertEquals(fieldSumDocIDs, sumDocIds);
+        assertEquals(fieldSumDocIDs, sumOrdToDocIds);
       }
     }
   }
@@ -1852,25 +1890,27 @@ public abstract class BaseKnnVectorsFormatTestCase extends BaseIndexFileFormatTe
 
     ByteVectorValues byteVectors = leafReader.getByteVectorValues("byte");
     assertNotNull(byteVectors);
-    assertEquals(0, byteVectors.nextDoc());
-    assertArrayEquals(new byte[] {42}, byteVectors.vectorValue());
-    assertEquals(1, byteVectors.nextDoc());
-    assertArrayEquals(new byte[] {42}, byteVectors.vectorValue());
-    assertEquals(DocIdSetIterator.NO_MORE_DOCS, byteVectors.nextDoc());
+    KnnVectorValues.DocIndexIterator iter = byteVectors.iterator();
+    assertEquals(0, iter.nextDoc());
+    assertArrayEquals(new byte[] {42}, byteVectors.vectorValue(0));
+    assertEquals(1, iter.nextDoc());
+    assertArrayEquals(new byte[] {42}, byteVectors.vectorValue(1));
+    assertEquals(DocIdSetIterator.NO_MORE_DOCS, iter.nextDoc());
 
     FloatVectorValues floatVectors = leafReader.getFloatVectorValues("float");
     assertNotNull(floatVectors);
-    assertEquals(0, floatVectors.nextDoc());
-    float[] vector = floatVectors.vectorValue();
+    iter = floatVectors.iterator();
+    assertEquals(0, iter.nextDoc());
+    float[] vector = floatVectors.vectorValue(0);
     assertEquals(2, vector.length);
     assertEquals(1f, vector[0], 0f);
     assertEquals(2f, vector[1], 0f);
-    assertEquals(1, floatVectors.nextDoc());
-    vector = floatVectors.vectorValue();
+    assertEquals(1, iter.nextDoc());
+    vector = floatVectors.vectorValue(1);
     assertEquals(2, vector.length);
     assertEquals(1f, vector[0], 0f);
     assertEquals(2f, vector[1], 0f);
-    assertEquals(DocIdSetIterator.NO_MORE_DOCS, floatVectors.nextDoc());
+    assertEquals(DocIdSetIterator.NO_MORE_DOCS, iter.nextDoc());
 
     IOUtils.close(reader, w2, dir1, dir2);
   }
